package com.github.soniex2.tmcg.ym2612;

public class OperatorUnit {

	private static final int[] sineTable = new int[256];
	private static final int[] powerTable = new int[256];

	static {
		// sinTableBitCount = 8
		// attenuationFixedBitCount = 8
		for (int i = 0; i < 256; ++i) {
			/*
			 * Calculate the normalized phase value for the input into the sine
			 * table. Note that this is calculated as a normalized result from
			 * 0.0-1.0 where 0 is not reached, because the phase is calculated
			 * as if it was a 9-bit index with the LSB fixed to 1. This was done
			 * so that the sine table would be more accurate when it was
			 * "mirrored" to create the negative oscillation of the wave. It's
			 * also convenient we don't have to worry about a phase of 0,
			 * because 0 is an invalid input for a log function, which we need
			 * to use below.
			 */
			double phaseNormalized = ((double) ((i << 1) + 1) / 512);

			/*
			 * Calculate the pure sine value for the input. Note that we only
			 * build a sine table for a quarter of the full oscillation
			 * (0-PI/2), since the upper two bits of the full phase are
			 * extracted by the external circuit.
			 */
			double sinResultNormalized = Math.sin(phaseNormalized
					* (Math.PI / 2));

			/*
			 * Convert the sine result from a linear representation of volume,
			 * to a logarithmic representation of attenuation. The YM2612 stores
			 * values in the sine table in this form because logarithms simplify
			 * multiplication down to addition, and this allowed them to
			 * attenuate the sine result by the envelope generator output simply
			 * by adding the two numbers together.
			 */
			// double sinResultAsAttenuation = -Math.log(sinResultNormalized)/
			// Math.log(2.0);
			double sinResultAsAttenuation = -log2(sinResultNormalized);
			/*
			 * The division by log(2) is required because the log function is
			 * base 10, but the YM2612 uses a base 2 logarithmic value. Dividing
			 * the base 10 log result by log10(2) will convert the result to a
			 * base 2 logarithmic value, which can then be converted back to a
			 * linear value by a pow2 function. In other words:
			 * 2^(log10(x)/log10(2)) = 2^log2(x) = x
			 * 
			 * If there was a native log2() function provided we could use that
			 * instead.
			 */

			/*
			 * Convert the attenuation value to a rounded 12-bit result in 4.8
			 * fixed point format.
			 */
			int sinResult = (int) ((sinResultAsAttenuation * 256) + 0.5);

			// Write the result to the table
			sineTable[i] = sinResult;
		}

		// powTableBitCount = 8
		// powTableOutputBitCount = 11
		for (int i = 0; i < (1 << 8); ++i) {
			/*
			 * Normalize the current index to the range 0.0-1.0. Note that in
			 * this case, 0.0 is a value which is never actually reached, since
			 * we start from i+1. They only did this to keep the result to an
			 * 11-bit output. It probably would have been better to simply
			 * subtract 1 from every final number and have 1.0 as the input
			 * limit instead when building the table, so an input of 0 would
			 * output 0x7FF, but they didn't.
			 */
			double entryNormalized = (double) (i + 1) / (double) 256;

			// Calculate 2^-entryNormalized
			double resultNormalized = Math.pow(2, -entryNormalized);

			// Convert the normalized result to an 11-bit rounded result
			int result = (int) ((resultNormalized * 2048) + 0.5);

			// Write the result to the table
			powerTable[i] = result;
		}
	}

	protected static double log2(double d) {
		return Math.log10(d) / Math.log10(2.0);
	}

	private PhaseGenerator gen;

	// Code by Nemesis. Found here:
	// http://gendev.spritesmind.net/forum/viewtopic.php?t=386&postdays=0&postorder=asc&start=150
	public OperatorUnit() {
	}

	public int inversePow2(int i) {
		int shift = i >> 8;
		int index = i & 255;
		int entry = powerTable[index];
		int shifted = (entry << 2) >> shift;
		return shifted;
	}

	// For self feedback, opUnit = opUnitOut << 1.
	// If you have 2 opUnitOuts, just add them together and sum << 1
	public int doStuff(int opUnit, int envelope) {
		int phase = gen.get();
		gen.update();
		opUnit &= ((1 << 10) - 1);
		phase &= ((1 << 10) - 1);
		envelope &= ((1 << 10) - 1);
		phase = opUnit + phase;
		phase &= ((1 << 10) - 1);
		boolean sign = (phase & (1 << 9)) != 0;
		phase &= ((1 << 9) - 1); // half-phase
		if ((phase & (1 << 8)) != 0) {
			phase = (~phase) & 256;
		} else {
			phase = phase & 256;
		}
		int attenuation = sineTable[phase];
		attenuation += envelope << 2;
		int out = inversePow2(attenuation);
		if (sign) {
			out ^= (1 << 14) - 1;
		}
		return out;
	}

	public void setPhaseGenerator(PhaseGenerator gen) {
		this.gen = gen;
	}

	public PhaseGenerator getPhaseGenerator() {
		return gen;
	}
}
